-------------------------------------------------------------------------------
-- Localization
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("SimpleSelfRebuff")
local BS = AceLibrary("Babble-Spell-2.2")

L:RegisterTranslations("enUS", function() return {} end)

-------------------------------------------------------------------------------
-- Aliases & constants
-------------------------------------------------------------------------------

local TEH_BUTTON_NAME = "SimpleSelfRebuffButton"

local trackings = {}
do
	local function initTrackings(...)
		for i=1,select('#', ...) do
			local t = select(i, ...)			
			trackings[BS:GetSpellIcon(t)] = BS[t]
		end
	end
	initTrackings(
		'Track Beasts',     'Track Demons', 'Track Dragonkin',
		'Track Elementals', 'Track Giants', 'Track Hidden',
		'Track Humanoids',  'Track Undead', 'Find Herbs',
		'Find Minerals',    'Find Treasure'
	)
end

local activationCheckEvents = {
	'PLAYER_DEAD', 'PLAYER_UNGHOST', 'PLAYER_ALIVE',
	'PLAYER_UPDATE_RESTING',
	'PLAYER_REGEN_DISABLED', 'PLAYER_REGEN_ENABLED',
	'PLAYER_CONTROL_GAINED', 'PLAYER_CONTROL_LOST',
}

-------------------------------------------------------------------------------
-- Locals
-------------------------------------------------------------------------------

local registeredBuffs = {}
local categories = {}
local availableBuffs = {}
local actualBuffNames = {}
local actualBuffExpirations = {}

local buffToCast
local castButton
local spellScanned

-- local englishSpells
local active

local db

-------------------------------------------------------------------------------
-- Externals
-------------------------------------------------------------------------------

local Tablet = AceLibrary('Tablet-2.0')
 
-------------------------------------------------------------------------------
-- Addon declaration
-------------------------------------------------------------------------------

SimpleSelfRebuff = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "FuBarPlugin-2.0")
local SimpleSelfRebuff = SimpleSelfRebuff

SimpleSelfRebuff.options = {	
	type = 'group',
	args = {
		buffHeader = {
			type = 'header',
			name = ' ',
			order = 199,
		}
	}
}
	
do
	local m = SimpleSelfRebuff
	m.registeredBuffs = registeredBuffs
	m.categories = categories
	m.availableBuffs = availableBuffs
	m.actualBuffNames = actualBuffNames
	m.actualBuffExpirations = actualBuffExpirations	
	m.englishSpells = englishSpells
end

-------------------------------------------------------------------------------
-- FuBar plugin settings
-------------------------------------------------------------------------------

SimpleSelfRebuff.hasNoColor = true
SimpleSelfRebuff.hideWithoutStandby = true
SimpleSelfRebuff.OnMenuRequest = SimpleSelfRebuff.options

-------------------------------------------------------------------------------
-- Helpers
-------------------------------------------------------------------------------

local function error(msg, ...)
	if select('#', ...) > 1 then
		msg = msg:format(...)
	end
	return _G.error(msg)
end

local function erase(...)
	for i=1,select('#', ...) do
		local t = select(i, ...)
		if type(t) == 'table' then
			for k in pairs(t) do
				t[k] = nil 
			end
		end
	end
end

local function colorize(text, r, g, b)
	if type(r) == 'table' and r.r then
		return colorize(msg, r.r, r.g, r.b)
	else
		return ("|cff%02x%02x%02x%s|r"):format(msg, r, g, b)
	end
end

-------------------------------------------------------------------------------
-- Initializing
-------------------------------------------------------------------------------

-- Addon setup
function SimpleSelfRebuff:OnInitialize()
	self.revision = tonumber(string.sub("$Revision: 41933 $", 12, -3))
	self.version = "r" .. SimpleSelfRebuff.revision
	self.date = string.sub("$Date: 2007-06-27 18:16:32 -0400 (Wed, 27 Jun 2007) $", 8, 17)

	self:RegisterDB("SimpleSelfRebuffDB")
	self:RegisterDefaults('profile', {
		categories = {},
	})

	db = self.db.profile

	self:RegisterChatCommand({"/ssr", "/simpleselfrebuff"}, self.options)
	
	if AceLibrary:HasInstance('AceDebug-2.0') then
		AceLibrary('AceDebug-2.0'):embed(SimpleSelfRebuff)
		SimpleSelfRebuff:SetDebugging(UnitName('player') == 'Adirelle')
	else
		function SimpleSelfRebuff.Debug() end
	end
		
end

function SimpleSelfRebuff:OnEnable(first)
	if first then
		self:ProcessRegistry()
	end

	self:RegisterBucketEvent('SPELLS_CHANGED', 0.1, 'ScanSpellbook')
	self:RegisterBucketEvent(activationCheckEvents, 0.1, 'CheckActivation')
	self:RegisterBucketEvent('PLAYER_AURAS_CHANGED', 0.2)
	self:RegisterBucketEvent('SimpleSelfRebuff_UpdateDisplay', 0.1, 'UpdateDisplay')
	
	self:ScanSpellbook()
	self:CheckActivation()
end

function SimpleSelfRebuff:PLAYER_AURAS_CHANGED()
	self:CheckActivation()
	if active then
		self:ScanPlayerAuras()
	end
end

-------------------------------------------------------------------------------
-- Buff registry
-------------------------------------------------------------------------------

do
	local buffFuncs = {}

	function SimpleSelfRebuff:RegisterBuffSetup(func)
		if type(func) == "function" then
			tinsert(buffFuncs, func)
		else
			error("Arg #1 to RegisterBuffs must be a function")
		end
	end
	
	local function registerCallback(...)
		return SimpleSelfRebuff:RegisterBuff(...)
	end

	function SimpleSelfRebuff:ProcessRegistry()
		if not buffFuncs then
			return
		end
		
		local selfClassName = select(2, UnitClass('player'))
		for i,func in ipairs(buffFuncs) do
			func(registerCallback, BS, selfClassName)
		end
		buffFuncs = nil

		-- Build categories
		for buffName, buffInfo in pairs(registeredBuffs) do
			if buffInfo.recast then
				local category = buffInfo.category
				if not categories[category] then
					categories[category] = {}
				end
				if not availableBuffs[category] then
					availableBuffs[category] = {}
				end
				tinsert(categories[category], buffName)
			end
		end
		
		-- Build categories options
		local opts = self.options.args
		for category, buffList in pairs(categories) do
			local category, buffList = category, buffList
			if #buffList == 1 then
				opts[category] = {
					type = 'toggle',
					name = category,
					desc = category,
					get = function()
						return db.categories[category] == category
					end,
					set = function(enable)
						db.categories[category] = enable and category or nil
						self:CheckRebuff()
					end,
					hidden = function()
						return not next(availableBuffs[category])
					end,
					order = 200,
				}
			else
				opts[category] = {
					type = 'text',
					name = category,
					desc = category,
					get = function()
						return db.categories[category]
					end,
					set = function(buffName)
						db.categories[category] = buffName
						self:CheckRebuff()
					end,
					validate = availableBuffs[category],
					hidden = function()
						return not next(availableBuffs[category])
					end,
					order = 200,
				}
			end
		end
	end
	
end

function SimpleSelfRebuff:RegisterBuff(name, category, recast)
	if registeredBuffs[name] then
		error("%q already registered", name)
	end
	registeredBuffs[name] = {
		name = name,
		category = category or name,
		recast = recast == nil or recast and true,
	}
end

-------------------------------------------------------------------------------
-- Activation
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckActivation()
	local shouldActivate = spellScanned and not (UnitIsDeadOrGhost('player') or InCombatLockdown() or IsResting()
	  or IsMounted() or IsFlying() or IsStealthed())

	if shouldActivate and not active then
		active = true
		self:Debug('Activated')
		
		self:ScheduleRepeatingEvent('SimpleSelfRebuff_CheckRebuff', self.CheckRebuff, 1.5, self)
		self:ScanPlayerAuras()
		self:CheckRebuff()
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
		
	elseif not shouldActivate and active then
		active = nil
		self:Debug('Disactivated')
		
		erase(actualBuffNames, actualBuffExpirations)

		self:CleanupPendingCast()		
		self:CancelScheduledEvent('SimpleSelfRebuff_CheckRebuff')
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

-------------------------------------------------------------------------------
-- Core
-------------------------------------------------------------------------------

function SimpleSelfRebuff:CheckRebuff()

	local newBuffToCast = nil	
	local shortestBuffName, shortestBuffDuration
	
	for category in pairs(categories) do
		local expectedBuff = db.categories[category] 
		
		-- Always check cooldown
		local inCooldown
		if expectedBuff then
			local start, duration = GetSpellCooldown(expectedBuff)
			if start and start + duration > GetTime() then
				inCooldown = true
			end
		end
		
		-- Is we can cast the spell, check it further
		if expectedBuff and not inCooldown and IsUsableSpell(expectedBuff) then			
			local actualBuff = actualBuffNames[category] 
			
			-- Check if it is missing
			if not actualBuff then
				newBuffToCast = expectedBuff
				self:Debug('Missing buff: %q', expectedBuff)
				break
				
			-- Check remaning time
			elseif actualBuff == expectedBuff and actualBuffExpirations[category] then
				local timeLeft = actualBuffExpirations[category] - GetTime()
				if timeLeft > 0 then
					if not shortestBuffDuration or timeLeft < shortestBuffDuration then
						shortestBuffDuration, shortestBuffName = timeLeft, actualBuff
					end
				end
			end
			
		end
	end
	
	if not newBuffToCast and shortestBuffName and shortestBuffDuration < 30 then
		newBuffToCast = shortestBuffName
		self:Debug('Buff needs refresh: %q (%q)', newBuffToCast, shortestBuffDuration)
	end

	if newBuffToCast ~= buffToCast then
		
		if newBuffToCast then
			
			if not castButton then
				castButton = CreateFrame("Button", TEH_BUTTON_NAME, UIParent, "SecureActionButtonTemplate")
				castButton:RegisterForClicks("LeftButtonUp")
				castButton:SetPoint("CENTER",0,0)
				castButton:Hide()
	
				castButton:SetAttribute('type', 'spell')
				castButton:SetAttribute('target', 'player')
			end
			
			buffToCast = newBuffToCast
			self:Debug('Binding mousewheel to %q', buffToCast)
			castButton:SetAttribute('spell', buffToCast)
			
			SetOverrideBindingClick(castButton, true, "MOUSEWHEELUP", TEH_BUTTON_NAME, "LeftButton")
			SetOverrideBindingClick(castButton, true, "MOUSEWHEELDOWN", TEH_BUTTON_NAME, "LeftButton")	

		else
			self:CleanupPendingCast()
		end
		
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end

end

function SimpleSelfRebuff:CleanupPendingCast()
	if buffToCast then
		self:Debug('Unbinding mousewheel')
		buffToCast = nil
		if castButton then
			ClearOverrideBindings(castButton)
		end
		self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
	end
end

-------------------------------------------------------------------------------
-- Spellbook handling
-------------------------------------------------------------------------------

--[[
englishSpells = setmetatable({}, {
	__index = function(self, key) 
		if BS:HasReverseTranslation(key) then
			rawset(self, key, BS:GetReverseTranslation(key))
		else
			rawset(self, key, key)
		end
		return self[key]
	end,
	__newindex = function() end,
})
--]]

function SimpleSelfRebuff:ScanSpellbook() 
	for k,v in pairs(availableBuffs) do
		erase(v)
	end
	local i = 1
	while true do
		local buffName = GetSpellName(i, BOOKTYPE_SPELL)
		if not buffName then
			break
		end
		local buffInfo = registeredBuffs[buffName]
		if buffInfo then
			availableBuffs[buffInfo.category][buffName] = buffName
		end
		spellScanned = true
		i = i + 1
	end
	self:CheckActivation()
end

-------------------------------------------------------------------------------
-- Player buff status tracking
-------------------------------------------------------------------------------

do
	local seen = {}
	
	function SimpleSelfRebuff:ScanPlayerAuras()
		erase(seen)
		local dirty
		self:Debug('ScanPlayerAuras')
		
		for i = 1, 32 do
			local index = GetPlayerBuff(i, "HELPFUL")
			buffName = index and GetPlayerBuffName(index)		
			if buffName and registeredBuffs[buffName] then
				local category = registeredBuffs[buffName].category
				seen[category] = true
				if not actualBuffNames[category] then
					actualBuffNames[category] = buffName
					dirty = true
					self:Debug('New buff: %q', buffName)
				end
				local timeLeft = GetPlayerBuffTimeLeft(index)
				if timeLeft then
					actualBuffExpirations[category] = GetTime() + timeLeft		
				else
					actualBuffExpirations[category] = nil
				end
			end
		end
		
		local trackingTexture = GetTrackingTexture()
		local trackingName = trackingTexture and trackings[trackingTexture]
		if trackingName and registeredBuffs[trackingName] then
			local category = registeredBuffs[trackingName].category			
			seen[category] = true
			if actualBuffNames[category] ~= trackingName then
				actualBuffNames[category] = trackingName
				self:Debug('New tracking: %q', buffName)
				dirty = true
			end
		end
		
		for category, actualBuff in pairs(actualBuffNames) do
			if not seen[category] then
				actualBuffNames[category] = nil
				actualBuffExpirations[category] = nil
				self:Debug('Faded out: %q', actualBuff)
				dirty = true
			end
		end
		
		if dirty then
			self:TriggerEvent('SimpleSelfRebuff_UpdateDisplay')
			self:Debug('Auras changed - checking rebuff')
			self:CheckRebuff()
		end
		
	end
	
end

-------------------------------------------------------------------------------
-- FuBar Plugin info
-------------------------------------------------------------------------------

function SimpleSelfRebuff:OnTextUpdate()
	if active then
		self:SetText('SSR')
	else
		self:SetText('|cff777777SSR|r')
	end
end

function SimpleSelfRebuff:OnTooltipUpdate()
	local cat = Tablet:AddCategory("columns", 2)
	for category in pairs(categories) do
		if next(availableBuffs[category]) then
			local expectedBuff = db.categories[category]
			local actualBuff = actualBuffNames[category]
			if actualBuff then
				if actualBuff == expectedBuff then
					cat:AddLine('text', category, 'text2', actualBuff, 'text2R', 0.3, 'text2G', 0.8, 'text2B', 0.2)
				else
					cat:AddLine('text', category, 'text2', actualBuff, 'text2R', 0.8, 'text2G', 0.8, 'text2B', 0.3)
				end
			else
				cat:AddLine('text', category, 'text2', expectedBuff, 'text2R', 0.8, 'text2G', 0.3, 'text2B', 0.3)
			end
		end
	end

end


